import ctypes
import io
import os

import PIL.Image as Image

from serbiacard.card_additional import (
    EID_CERTIFICATE,
    EID_DOCUMENT_DATA,
    EID_FIXED_PERSONAL_DATA,
    EID_PORTRAIT,
    EID_VARIABLE_PERSONAL_DATA,
)


class CarCard:
    pass


class IDCard:
    """With all data fields from ID with .run() to
    start reader, collect from chip, end reader and store information in this object.

    More information available through the CelikAPI documentation.
    Vise informacija u CelikAPI dokumentaciji.

        docRegNo: ID Number (Broj Licne)
        documentType: Information
        issuingDate: Information
        expiryDate: Information
        issuingAuthority: Information
        documentSerialNumber: Information
        chipSerialNumber: Information

        personalNumber: Information
        surname (or lastName): Information
        givenName (or firstName): Information
        parentGivenName: Information
        sex: Information
        placeOfBirth: Information
        stateOfBirth: Information
        dateOfBirth: Information
        communityOfBirth: Information
        statusOfForeigner: Information
        nationalityFull: Information

        state: Information
        community: Information
        place: Information
        street: Information
        houseNumber: Information
        houseLetter: Information
        entrance: Information
        floor: Information
        apartmentNumber: Information
        addressDate: Information
        addressLabel: Information

        portrait: Information.BytesIO()

        certificate: Information.BytesIO()
    """

    def __init__(self):
        self.docRegNo = None
        self.documentType = None
        self.issuingDate = None
        self.expiryDate = None
        self.issuingAuthority = None
        self.documentSerialNumber = None
        self.chipSerialNumber = None

        self.personalNumber = None
        self.surname = None
        self.givenName = None
        self.parentGivenName = None
        self.sex = None
        self.placeOfBirth = None
        self.stateOfBirth = None
        self.dateOfBirth = None
        self.communityOfBirth = None
        self.statusOfForeigner = None
        self.nationalityFull = None

        self.state = None
        self.community = None
        self.place = None
        self.street = None
        self.houseNumber = None
        self.houseLetter = None
        self.entrance = None
        self.floor = None
        self.apartmentNumber = None
        self.addressDate = None
        self.addressLabel = None

        self.portrait = io.BytesIO()

        self.certificate = None

        # Flag for checking if ID is read correctly
        # if read correctly > True, if not then False, default False
        self.idCheck = False
        self.startup_result = None

    def run(self):
        """Full complete: Completes reading, filling all information and closing the reader.

        This method does all tasks, finishes them completely including starting the reader,
        reading and filling the data inside the object values and closing the reader.
        """
        self.readerStart()
        self.readData()
        self.readerEnd()
        self.validate()

        self._encodeFields()

    def __repr__(self):
        information = self.__dict__
        del information["celikApi"]
        information["portrait"] = "Object portrait"
        information["certificate"] = "Object certificate"
        return f"{information}"

    def readerStart(self):
        # Getting the directory of the current script
        script_dir = os.path.dirname(os.path.abspath(__file__))

        # celik_path is the dll path of the API
        celik_path = os.path.join(script_dir, "CelikApi.dll")
        self.celikApi = ctypes.CDLL(celik_path)

        self.startup_result = self.celikApi.EidStartup(3)
        self.idCheck = True if not self.celikApi.EidBeginRead(ctypes.c_char(), 0) else False

    def readerEnd(self):
        self.celikApi.EidEndRead()
        self.celikApi.EidCleanup()

    def readData(self):
        # read all the fields
        self.setDocumentData()
        self.setFixedPersonalData()
        self.setVariablePersonalData()
        self.setPortrait()
        self.setCertificate()

    def setDocumentData(self):
        INTP = ctypes.POINTER(ctypes.c_int)
        obj = EID_DOCUMENT_DATA()
        addr = ctypes.addressof(obj)
        pointer_addr = ctypes.cast(addr, INTP)
        self.celikApi.EidReadDocumentData(pointer_addr)

        self.docRegNo = obj.docRegNo
        self.documentType = obj.documentType
        self.issuingDate = obj.issuingDate
        self.expiryDate = obj.expiryDate
        self.issuingAuthority = obj.issuingAuthority
        self.documentSerialNumber = obj.documentSerialNumber
        self.chipSerialNumber = obj.chipSerialNumber

        return True

    def setFixedPersonalData(self):
        INTP = ctypes.POINTER(ctypes.c_int)
        obj = EID_FIXED_PERSONAL_DATA()
        addr = ctypes.addressof(obj)
        pointer_addr = ctypes.cast(addr, INTP)
        self.celikApi.EidReadFixedPersonalData(pointer_addr)

        self.personalNumber = obj.personalNumber
        self.surname = obj.surname
        self.givenName = obj.givenName

        # additional
        self.firstName = self.givenName
        self.lastName = self.surname

        self.parentGivenName = obj.parentGivenName
        self.sex = obj.sex
        self.placeOfBirth = obj.placeOfBirth
        self.stateOfBirth = obj.stateOfBirth
        self.dateOfBirth = obj.dateOfBirth
        self.communityOfBirth = obj.communityOfBirth
        self.statusOfForeigner = obj.statusOfForeigner
        self.nationalityFull = obj.nationalityFull

        return True

    def setVariablePersonalData(self):
        INTP = ctypes.POINTER(ctypes.c_int)
        obj = EID_VARIABLE_PERSONAL_DATA()
        addr = ctypes.addressof(obj)
        pointer_addr = ctypes.cast(addr, INTP)
        self.celikApi.EidReadVariablePersonalData(pointer_addr)

        self.state = obj.state
        self.community = obj.community
        self.place = obj.place
        self.street = obj.street
        self.houseNumber = obj.houseNumber
        self.houseLetter = obj.houseLetter
        self.entrance = obj.entrance
        self.floor = obj.floor
        self.apartmentNumber = obj.apartmentNumber
        self.addressDate = obj.addressDate
        self.addressLabel = obj.addressLabel

        return True

    def setPortrait(self):
        INTP = ctypes.POINTER(ctypes.c_int)
        obj = EID_PORTRAIT()
        addr = ctypes.addressof(obj)
        pointer_addr = ctypes.cast(addr, INTP)
        self.celikApi.EidReadPortrait(pointer_addr)

        self.portrait = io.BytesIO(obj.portrait)

        return True

    # TODO: multiple types of certificate available
    def setCertificate(self, cert_choice=3):
        INTP = ctypes.POINTER(ctypes.c_int)
        obj = EID_CERTIFICATE()
        addr = ctypes.addressof(obj)
        pointer_addr = ctypes.cast(addr, INTP)
        self.celikApi.EidReadCertificate(pointer_addr)

        self.certificate = obj.certificate

        return True

    def showPortrait(self) -> None:
        image = Image.open(self.portrait)
        image.show()

    def _encodeFields(self):
        for attribute, value in self.__dict__.items():
            if isinstance(value, bytes):
                self.__dict__[attribute] = value.decode("UTF-8")
                # setattr(self, attribute, value.decode('UTF-8'))

    @property
    def is_valid(self):
        return self.idCheck

    def validate(self):
        """Validate if the ID is loaded correctly.

        Validate ID card if it is read correctly,
        raises and Exception if there is an Error in read
        returns True if there's no error

        Returns:
            True: everything's ok
            bool

        Raises:
            Exception: Error if not read.
        """
        if not self.idCheck:
            raise Exception("ID is not loaded correctly. Code:", self.startup_result)
        return True

    def __getitem__(self, attribute):
        return self.__dict__[attribute]


if __name__ == "__main__":
    card = IDCard()
    card.run()
    print(card)



import ctypes


class EID_DOCUMENT_DATA(ctypes.Structure):
    _fields_ = [
        ("docRegNo", ctypes.c_char * 9),
        ("docRegNoSize", ctypes.c_int),
        ("documentType", ctypes.c_char * 2),
        ("documentTypeSize", ctypes.c_int),
        ("issuingDate", ctypes.c_char * 10),
        ("issuingDateSize", ctypes.c_int),
        ("expiryDate", ctypes.c_char * 10),
        ("expiryDateSize", ctypes.c_int),
        ("issuingAuthority", ctypes.c_char * 100),
        ("issuingAuthoritySize", ctypes.c_int),
        ("documentSerialNumber", ctypes.c_char * 10),
        ("documentSerialNumberSize", ctypes.c_int),
        ("chipSerialNumber", ctypes.c_char * 14),
        ("chipSerialNumberSize", ctypes.c_int),
    ]


class EID_FIXED_PERSONAL_DATA(ctypes.Structure):
    _fields_ = [
        ("personalNumber", ctypes.c_char * 13),
        ("personalNumberSize", ctypes.c_int),
        ("surname", ctypes.c_char * 200),
        ("surnameSize", ctypes.c_int),
        ("givenName", ctypes.c_char * 200),
        ("givenNameSize", ctypes.c_int),
        ("parentGivenName", ctypes.c_char * 200),
        ("parentGivenNameSize", ctypes.c_int),
        ("sex", ctypes.c_char * 2),
        ("sexSize", ctypes.c_int),
        ("placeOfBirth", ctypes.c_char * 200),
        ("placeOfBirthSize", ctypes.c_int),
        ("stateOfBirth", ctypes.c_char * 200),
        ("stateOfBirthSize", ctypes.c_int),
        ("dateOfBirth", ctypes.c_char * 10),
        ("dateOfBirthSize", ctypes.c_int),
        ("communityOfBirth", ctypes.c_char * 200),
        ("communityOfBirthSize", ctypes.c_int),
        ("statusOfForeigner", ctypes.c_char * 200),
        ("statusOfForeignerSize", ctypes.c_int),
        ("nationalityFull", ctypes.c_char * 200),
        ("nationalityFullSize", ctypes.c_int),
    ]


class EID_VARIABLE_PERSONAL_DATA(ctypes.Structure):
    _fields_ = [
        ("state", ctypes.c_char * 100),
        ("stateSize", ctypes.c_int),
        ("community", ctypes.c_char * 200),
        ("communitySize", ctypes.c_int),
        ("place", ctypes.c_char * 200),
        ("placeSize", ctypes.c_int),
        ("street", ctypes.c_char * 200),
        ("streetSize", ctypes.c_int),
        ("houseNumber", ctypes.c_char * 20),
        ("houseNumberSize", ctypes.c_int),
        ("houseLetter", ctypes.c_char * 8),
        ("houseLetterSize", ctypes.c_int),
        ("entrance", ctypes.c_char * 10),
        ("entranceSize", ctypes.c_int),
        ("floor", ctypes.c_char * 6),
        ("floorSize", ctypes.c_int),
        ("apartmentNumber", ctypes.c_char * 12),
        ("apartmentNumberSize", ctypes.c_int),
        ("addressDate", ctypes.c_char * 10),
        ("addressDateSize", ctypes.c_int),
        ("addressLabel", ctypes.c_char * 60),
        ("addressLabelSize", ctypes.c_int),
    ]


class EID_PORTRAIT(ctypes.Structure):
    _fields_ = [("portrait", ctypes.c_byte * 7700), ("portraitSize", ctypes.c_int)]


class EID_CERTIFICATE(ctypes.Structure):
    _fields_ = [("certificate", ctypes.c_byte * 2048), ("certificateSize", ctypes.c_int)]
